@@ -20,6 +20,7 @@
 #include "application.h"
 #include "config.h"
 #include "gwater-xcb.h"
+#include "sn.h"
 #include "xcb-utils.h"
 
 #include <gio/gio.h>
@@ -37,6 +38,7 @@ struct _XEmbedSNIApplication
 	GApplication parent;
 	uint8_t damageEventBase;
 	GWaterXcbSource *src;
+	int default_screen;
 	xcb_window_t selection;
 	GHashTable *damageWatches;
 	GHashTable *proxies;
@@ -128,8 +130,14 @@ static void xembed_sni_application_dock(XEmbedSNIApplication *self, xcb_window_t
 
 	if (xembed_sni_application_add_damage_watch(self, winId))
 	{
-		//        g_hash_table_insert(self->proxies,GUINT_TO_POINTER(winID),);
-		//        m_proxies[winId] = new SNIProxy(winId, this);
+		g_autofree char *id = g_strdup_printf("0x%x", winId);
+		StatusNotifierItem *item =
+		    status_notifier_item_new_from_xcb_window(id,
+		                                             SN_CATEGORY_APPLICATION,
+		                                             g_water_xcb_source_get_connection(
+		                                                 self->src),
+		                                             winId);
+		g_hash_table_insert(self->proxies, GUINT_TO_POINTER(winId), item);
 	}
 }
 
@@ -144,7 +152,10 @@ static void xembed_sni_application_undock(XEmbedSNIApplication *self, xcb_window
 	g_hash_table_remove(self->proxies, GUINT_TO_POINTER(winId));
 }
 
-/*XCB event filter*/
+/*XCB event filter
+ * TODO: Baloon messages
+ * TODO: Any strange tray types
+ */
 
 static bool xcb_event_filter(xcb_generic_event_t *event, XEmbedSNIApplication *self)
 {
@@ -159,8 +170,32 @@ static bool xcb_event_filter(xcb_generic_event_t *event, XEmbedSNIApplication *s
 			case SYSTEM_TRAY_REQUEST_DOCK:
 				xembed_sni_application_dock(self, ce->data.data32[2]);
 				//                    return GDK_FILTER_REMOVE;
+				break;
+			case SYSTEM_TRAY_BEGIN_MESSAGE:
+				/* If a Begin Message event. look up the tray icon and execute it.
+				 */
+				//                balloon_message_begin_event(tr,
+				//                (XClientMessageEvent *)xev); return
+				//                GDK_FILTER_REMOVE;
+				break;
+
+			case SYSTEM_TRAY_CANCEL_MESSAGE:
+				/* If a Cancel Message event. look up the tray icon and execute it.
+				 */
+				//                balloon_message_cancel_event(tr,
+				//                (XClientMessageEvent *)xev); return
+				//                GDK_FILTER_REMOVE;
+				break;
 			}
 		}
+
+		else if (ce->type == a_NET_SYSTEM_TRAY_MESSAGE_DATA)
+		{
+			/* Client message of type _NET_SYSTEM_TRAY_MESSAGE_DATA.
+			 * Look up the tray icon and execute it. */
+			//            balloon_message_data_event(tr, (XClientMessageEvent *)xev);
+			//            return GDK_FILTER_REMOVE;
+		}
 	}
 	else if (responseType == XCB_UNMAP_NOTIFY)
 	{
@@ -172,6 +207,10 @@ static bool xcb_event_filter(xcb_generic_event_t *event, XEmbedSNIApplication *s
 	}
 	else if (responseType == XCB_DESTROY_NOTIFY)
 	{
+		/* Look for DestroyNotify events on tray icon windows and update state.
+		 * We do it this way rather than with a "plug_removed" event because delivery
+		 * of plug_removed events is observed to be unreliable if the client
+		 * disconnects within less than 10 ms. */
 		const xcb_window_t destroyedWId = ((xcb_destroy_notify_event_t *)event)->window;
 		if (g_hash_table_contains(self->proxies, GUINT_TO_POINTER(destroyedWId)))
 		{
@@ -187,7 +226,7 @@ static bool xcb_event_filter(xcb_generic_event_t *event, XEmbedSNIApplication *s
 		{
 			//            sniProx->update();
 			xcb_damage_subtract(g_water_xcb_source_get_connection(self->src),
-			                    GDK_POINTER_TO_XID(
+			                    GPOINTER_TO_UINT(
 			                        g_hash_table_lookup(self->damageWatches,
 			                                            GUINT_TO_POINTER(damagedWId))),
 			                    XCB_NONE,
@@ -195,196 +234,24 @@ static bool xcb_event_filter(xcb_generic_event_t *event, XEmbedSNIApplication *s
 		}
 	}
 
-	//        if (event->type == DestroyNotify)
-	//        {
-	//            /* Look for DestroyNotify events on tray icon windows and update state.
-	//             * We do it this way rather than with a "plug_removed" event because delivery
-	//             * of plug_removed events is observed to be unreliable if the client
-	//             * disconnects within less than 10 ms. */
-	//            XDestroyWindowEvent *xev_destroy = (XDestroyWindowEvent *)xev;
-	//            TrayClient *tc                   = client_lookup(tr, xev_destroy->window);
-	//            if (tc != NULL)
-	//                client_delete(tr, tc, TRUE, TRUE);
-	//        }
-
-	//        else if (xev->type == ClientMessage)
-	//        {
-	//            if (xev->xclient.message_type == a_NET_SYSTEM_TRAY_OPCODE)
-	//            {
-	//                /* Client message of type _NET_SYSTEM_TRAY_OPCODE.
-	//                 * Dispatch on the request. */
-	//                switch (xev->xclient.data.l[1])
-	//                {
-	//                case SYSTEM_TRAY_REQUEST_DOCK:
-	//                    /* If a Request Dock event on the invisible window, which is holding
-	//                     * the manager selection, execute it. */
-	//                    if (xev->xclient.window == tr->invisible_window)
-	//                    {
-	//                        trayclient_request_dock(tr, (XClientMessageEvent *)xev);
-	//                        return GDK_FILTER_REMOVE;
-	//                    }
-	//                    break;
-
-	//                case SYSTEM_TRAY_BEGIN_MESSAGE:
-	//                    /* If a Begin Message event. look up the tray icon and execute it.
-	//                     */
-	//                    balloon_message_begin_event(tr, (XClientMessageEvent *)xev);
-	//                    return GDK_FILTER_REMOVE;
-
-	//                case SYSTEM_TRAY_CANCEL_MESSAGE:
-	//                    /* If a Cancel Message event. look up the tray icon and execute it.
-	//                     */
-	//                    balloon_message_cancel_event(tr, (XClientMessageEvent *)xev);
-	//                    return GDK_FILTER_REMOVE;
-	//                }
-	//            }
-
-	//            else if (xev->xclient.message_type == a_NET_SYSTEM_TRAY_MESSAGE_DATA)
-	//            {
-	//                /* Client message of type _NET_SYSTEM_TRAY_MESSAGE_DATA.
-	//                 * Look up the tray icon and execute it. */
-	//                balloon_message_data_event(tr, (XClientMessageEvent *)xev);
-	//                return GDK_FILTER_REMOVE;
-	//            }
-	//        }
-
-	//        else if ((xev->type == SelectionClear) && (xev->xclient.window ==
-	//        tr->invisible_window))
-	//        {
-	//            /* Look for SelectionClear events on the invisible window, which is holding
-	//            the
-	//             * manager selection.
-	//             * This should not happen. */
-	//            tray_unmanage_selection(tr);
-	//        }
-	return GDK_FILTER_CONTINUE;
+	else if (responseType == XCB_SELECTION_CLEAR)
+	{
+		const xcb_window_t damagedWId = ((xcb_selection_clear_event_t *)event)->owner;
+		/* Look for SelectionClear events on the invisible window, which is holding
+		the
+		 * manager selection.
+		 * This should not happen. */
+		if (damagedWId == self->selection)
+			g_application_quit(G_APPLICATION(self));
+	}
+	return G_SOURCE_CONTINUE;
 }
 
-/* GDK event filter. */
-// static GdkFilterReturn xembed_sni_event_filter(XEvent *xev, GdkEvent *event,
-//                                               XEmbedSNIApplication *self)
-//{
-//	// Try to convert XEvent to xcb_event_t
-//	xcb_generic_event_t *ev = (xcb_generic_event_t *)xev;
-//	//    xcb_connection_t *con      = gdk_x11_get_default_xcb_connection();
-//	const uint8_t responseType = ev->response_type;
-//	if (responseType == XCB_CLIENT_MESSAGE)
-//	{
-//		const xcb_client_message_event_t *ce = (xcb_client_message_event_t *)ev;
-//		if (ce->type == a_NET_SYSTEM_TRAY_OPCODE)
-//		{
-//			switch (ce->data.data32[1])
-//			{
-//			case SYSTEM_TRAY_REQUEST_DOCK:
-//				xembed_sni_application_dock(self, ce->data.data32[2]);
-//				return GDK_FILTER_REMOVE;
-//			}
-//		}
-//	}
-//	else if (responseType == XCB_UNMAP_NOTIFY)
-//	{
-//		const xcb_window_t unmappedWId = ((xcb_unmap_notify_event_t *)ev)->window;
-//		if (g_hash_table_contains(self->proxies, GUINT_TO_POINTER(unmappedWId)))
-//		{
-//			xembed_sni_application_undock(self, unmappedWId);
-//		}
-//	}
-//	else if (responseType == XCB_DESTROY_NOTIFY)
-//	{
-//		const xcb_window_t destroyedWId = ((xcb_destroy_notify_event_t *)ev)->window;
-//		if (g_hash_table_contains(self->proxies, GUINT_TO_POINTER(destroyedWId)))
-//		{
-//			xembed_sni_application_undock(self, destroyedWId);
-//		}
-//	}
-//	else if (responseType == self->damageEventBase + XCB_DAMAGE_NOTIFY)
-//	{
-//		const xcb_window_t damagedWId = ((xcb_damage_notify_event_t *)ev)->drawable;
-//		GObject *sniProxy =
-//		    G_OBJECT(g_hash_table_lookup(self->proxies, GUINT_TO_POINTER(damagedWId)));
-//		if (sniProxy)
-//		{
-//			//            sniProx->update();
-//			xcb_damage_subtract(gdk_x11_get_default_xcb_connection(),
-//			                    GDK_POINTER_TO_XID(
-//			                        g_hash_table_lookup(self->damageWatches,
-//			                                            GUINT_TO_POINTER(
-//			                                                damagedWId))),
-//			                    XCB_NONE,
-//			                    XCB_NONE);
-//		}
-//	}
-
-//    if (xev->type == DestroyNotify)
-//    {
-//        /* Look for DestroyNotify events on tray icon windows and update state.
-//         * We do it this way rather than with a "plug_removed" event because delivery
-//         * of plug_removed events is observed to be unreliable if the client
-//         * disconnects within less than 10 ms. */
-//        XDestroyWindowEvent *xev_destroy = (XDestroyWindowEvent *)xev;
-//        TrayClient *tc                   = client_lookup(tr, xev_destroy->window);
-//        if (tc != NULL)
-//            client_delete(tr, tc, TRUE, TRUE);
-//    }
-
-//    else if (xev->type == ClientMessage)
-//    {
-//        if (xev->xclient.message_type == a_NET_SYSTEM_TRAY_OPCODE)
-//        {
-//            /* Client message of type _NET_SYSTEM_TRAY_OPCODE.
-//             * Dispatch on the request. */
-//            switch (xev->xclient.data.l[1])
-//            {
-//            case SYSTEM_TRAY_REQUEST_DOCK:
-//                /* If a Request Dock event on the invisible window, which is holding
-//                 * the manager selection, execute it. */
-//                if (xev->xclient.window == tr->invisible_window)
-//                {
-//                    trayclient_request_dock(tr, (XClientMessageEvent *)xev);
-//                    return GDK_FILTER_REMOVE;
-//                }
-//                break;
-
-//            case SYSTEM_TRAY_BEGIN_MESSAGE:
-//                /* If a Begin Message event. look up the tray icon and execute it.
-//                 */
-//                balloon_message_begin_event(tr, (XClientMessageEvent *)xev);
-//                return GDK_FILTER_REMOVE;
-
-//            case SYSTEM_TRAY_CANCEL_MESSAGE:
-//                /* If a Cancel Message event. look up the tray icon and execute it.
-//                 */
-//                balloon_message_cancel_event(tr, (XClientMessageEvent *)xev);
-//                return GDK_FILTER_REMOVE;
-//            }
-//        }
-
-//        else if (xev->xclient.message_type == a_NET_SYSTEM_TRAY_MESSAGE_DATA)
-//        {
-//            /* Client message of type _NET_SYSTEM_TRAY_MESSAGE_DATA.
-//             * Look up the tray icon and execute it. */
-//            balloon_message_data_event(tr, (XClientMessageEvent *)xev);
-//            return GDK_FILTER_REMOVE;
-//        }
-//    }
-
-//    else if ((xev->type == SelectionClear) && (xev->xclient.window ==
-//    tr->invisible_window))
-//    {
-//        /* Look for SelectionClear events on the invisible window, which is holding the
-//         * manager selection.
-//         * This should not happen. */
-//        tray_unmanage_selection(tr);
-//    }
-
-//	return GDK_FILTER_CONTINUE;
-//}
-
 static void claim_systray_selection(XEmbedSNIApplication *self)
 {
 	xcb_connection_t *con = g_water_xcb_source_get_connection(self->src);
 	g_autofree char *selection_atom_name =
-	    g_strdup_printf("_NET_SYSTEM_TRAY_S%d", gdk_x11_get_default_screen());
+	    g_strdup_printf("_NET_SYSTEM_TRAY_S%d", self->default_screen);
 	xcb_atom_t selection_atom =
 	    (xcb_atom_t)xcb_atom_get_for_connection(g_water_xcb_source_get_connection(self->src),
 	                                            selection_atom_name);
@@ -407,7 +274,9 @@ static void claim_systray_selection(XEmbedSNIApplication *self)
 	uint32_t values[] = { true,
 		              XCB_EVENT_MASK_PROPERTY_CHANGE | XCB_EVENT_MASK_STRUCTURE_NOTIFY };
 	self->selection   = xcb_generate_id(con);
-	xcb_window_t root = (xcb_window_t)gdk_x11_get_default_root_xwindow();
+	xcb_window_t root =
+	    (xcb_window_t)xcb_get_root_for_connection(g_water_xcb_source_get_connection(self->src),
+	                                              self->default_screen);
 	xcb_create_window(con,
 	                  XCB_COPY_FROM_PARENT,
 	                  self->selection,
@@ -474,10 +343,10 @@ static void release_systray_selection(XEmbedSNIApplication *self)
 	{
 		xcb_connection_t *con = g_water_xcb_source_get_connection(self->src);
 		g_autofree char *selection_atom_name =
-		    g_strdup_printf("_NET_SYSTEM_TRAY_S%d", gdk_x11_get_default_screen());
+		    g_strdup_printf("_NET_SYSTEM_TRAY_S%d", self->default_screen);
 		xcb_atom_t selection_atom =
-		    (xcb_atom_t)gdk_x11_get_xatom_by_name_for_display(gdk_display_get_default(),
-		                                                      selection_atom_name);
+		    xcb_atom_get_for_connection(g_water_xcb_source_get_connection(self->src),
+		                                selection_atom_name);
 		g_autofree xcb_generic_error_t *err = NULL;
 		xcb_get_selection_owner_cookie_t sel_c =
 		    xcb_get_selection_owner(con, selection_atom);
@@ -515,7 +384,7 @@ static void xembed_sni_application_startup(GApplication *base)
 	// We only support X11
 	self->src = g_water_xcb_source_new(g_main_context_default(),
 	                                   "0",
-	                                   0,
+	                                   &self->default_screen,
 	                                   (GWaterXcbEventCallback)xcb_event_filter,
 	                                   self,
 	                                   NULL);
@@ -538,14 +407,11 @@ static void xembed_sni_application_startup(GApplication *base)
 	}
 	resolve_atoms(con);
 	claim_systray_selection(self);
-	/* Add GDK event filter. */
-	//	gdk_window_add_filter(NULL, (GdkFilterFunc)xembed_sni_event_filter, self);
 }
 
 static void xembed_sni_application_shutdown(GApplication *base)
 {
 	XEmbedSNIApplication *self = XEMBED_SNI_APPLICATION(base);
-	//	gdk_window_remove_filter(NULL, (GdkFilterFunc)xembed_sni_event_filter, self);
 	release_systray_selection(self);
 	G_APPLICATION_CLASS(xembed_sni_application_parent_class)
 	    ->shutdown((GApplication *)G_TYPE_CHECK_INSTANCE_CAST(base,
@@ -582,7 +448,7 @@ static void xembed_sni_app_finalize(GObject *object)
 	(*G_OBJECT_CLASS(xembed_sni_application_parent_class)->finalize)(object);
 }
 
-static void xembed_sni_app_set_property(GObject *object, guint prop_id, const GValue *value,
+static void xembed_sni_app_set_property(GObject *object, uint prop_id, const GValue *value,
                                         GParamSpec *pspec)
 {
 	XEmbedSNIApplication *app;
@@ -598,7 +464,7 @@ static void xembed_sni_app_set_property(GObject *object, guint prop_id, const GV
 	}
 }
 
-static void xembed_sni_app_get_property(GObject *object, guint prop_id, GValue *value,
+static void xembed_sni_app_get_property(GObject *object, uint prop_id, GValue *value,
                                         GParamSpec *pspec)
 {
 	XEmbedSNIApplication *app;
